/*
 * VIA VT 82c686[AB] high speed serial port enabler
 * Version 0.92
 * Copyright (c) 2000-2001 Juhani Rautiainen <jrauti@iki.fi>
 *
 * 0.92:
 *	Ported to 2.5/2.6 by Kingsly John
 *	- Corrected locking (no more cli() and sti())
 *	- New makefile
 *	
 * Can be freely distributed and used under the terms of the GNU GPL.
*/

#include <linux/module.h>
#include <linux/version.h>
#include <linux/init.h>
#include <linux/pci.h>

const unsigned short confindex=0x3F0,confdata=0x3F1;
const unsigned char spcidx=0xEE;

spinlock_t driver_lock = SPIN_LOCK_UNLOCKED;

static int __init viahss_init(void)
{
        struct pci_dev *pcidev = NULL;
        unsigned char confval,val;
        pcidev = pci_find_device (PCI_VENDOR_ID_VIA,PCI_DEVICE_ID_VIA_82C686,NULL);
        if (pcidev) {
                spin_lock_irq(&driver_lock);
                /* start config */
                pci_read_config_byte(pcidev,0x85,&confval);
                confval |= 0x2;
                pci_write_config_byte (pcidev, 0x85,confval);
                /* activate high speed bits */
                outb(spcidx,confindex); /* set index */
                val = (unsigned char) inb(confdata);
                val |= 0xC0;        /* both ports on high speed*/
                outb (spcidx,confindex);
                outb (val,confdata);
                /*stop config*/
                confval &= ~0x2;
                pci_write_config_byte(pcidev, 0x85, confval);
                spin_unlock_irq(&driver_lock);
                printk (KERN_INFO "VIA VT82C686[AB] serial port high speed enabled\n");
        }
        else {
                printk (KERN_INFO "Couldn't locate VIA chipset\n");
                return -ENODEV;
        }
        return 0;
}

static void __exit viahss_exit(void)
{
        struct pci_dev *pcidev = NULL;
        unsigned char confval,val;
        pcidev = pci_find_device (PCI_VENDOR_ID_VIA,PCI_DEVICE_ID_VIA_82C686,NULL);
        if (pcidev) {
                spin_lock_irq(&driver_lock);
                /* start config */
                pci_read_config_byte(pcidev,0x85,&confval);
                confval |= 0x2;
                pci_write_config_byte (pcidev, 0x85,confval);
                /* activate high speed bits */
                outb(spcidx,confindex); /* set index */
                val = (unsigned char) inb(confdata);
                val &= ~0xC0;        /* both ports off high speed*/
                outb (spcidx,confindex);
                outb (val,confdata);
                /*stop config*/
                confval &= ~0x2;
                pci_write_config_byte(pcidev, 0x85, confval);
                spin_unlock_irq(&driver_lock);
				printk (KERN_INFO "VIA VT82C686[AB] serial port high speed disabled\n");
        }
}

module_init(viahss_init);
module_exit(viahss_exit);
MODULE_DESCRIPTION("VIA VT82C686[AB] high speed serial port enabler");
MODULE_AUTHOR("Juhani Rautiainen <jrauti@iki.fi>");
MODULE_LICENSE("GPL");
