#ifndef __SOUND_VT1732_H
#define __SOUND_VT1712_H

/*
 *   ALSA driver for VT1732 (Envy24-II)
 *
 *	Copyright (c) 2006 Takashi Iwai <tiwai@suse.de>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 *
 */      

#include <sound/control.h>
#include <sound/rawmidi.h>
#include <sound/i2c.h>
#include <sound/pcm.h>



#define VT1732_DMA_INT_MASK		0x000
#define   VT1732_UART1_TX_INT_MASK	(1 << 15)
#define   VT1732_UART2_TX_INT_MASK	(1 << 14)
#define   VT1732_REC_XRUN_INT_MASK	(1 << 13)
#define   VT1732_PLAY_XRUN_INT_MASK	(1 << 12)
#define   VT1732_UART1_RX_INT_MASK	(1 << 11)
#define   VT1732_UART2_RX_INT_MASK	(1 << 10)
#define   VT1732_GPIO_INT_STAT		(1 << 8)
#define   VT1732_UART1_TX_INT_STAT	(1 << 7)
#define   VT1732_UART2_TX_INT_STAT	(1 << 6)
#define   VT1732_REC_XRUN_INT_STAT	(1 << 5)
#define   VT1732_PLAY_XRUN_INT_STAT	(1 << 4)
#define   VT1732_UART1_RX_INT_STAT	(1 << 3)
#define   VT1732_UART2_RX_INT_STAT	(1 << 2)
#define   VT1732_SPDF_REC_INT_STAT	(1 << 1)
#define   VT1732_SPDF_PLAY_INT_STAT	(1 << 0)

#define VT1732_LEGACY_CTRL		0x003
#define   VT1732_RESET			(1 << 7)
#define   VT1732_PCI_MASTER		(1 << 2)
#define   VT1732_PCI_SLAVE		(1 << 1)
#define   VT1732_NATIVE			(1 << 0)

#define VT1732_PLAY_INT_MASK		0x004
#define VT1732_PLAY_INT			0x008
#define VT1732_REC_INT_MASK		0x00c
#define VT1732_REC_INT			0x010
#define VT1732_PLAY_START		0x014
#define VT1732_PLAY_PAUSE		0x018
#define VT1732_REC_START		0x01c

#define VT1732_RATE_SELECT		0x020
#define   VT1732_PBCLK_OUTPUT		(1 << 7)
#define   VT1732_SPDIF_IN_MASTER	(1 << 4)
#define   VT1732_CODEC_RATE		(0xf << 0)
#define   VT1732_CODEC_RATE_48		0
#define   VT1732_CODEC_RATE_24		1
#define   VT1732_CODEC_RATE_12		2
#define   VT1732_CODEC_RATE_9		3
#define   VT1732_CODEC_RATE_32		4
#define   VT1732_CODEC_RATE_16		5
#define   VT1732_CODEC_RATE_8		6
#define   VT1732_CODEC_RATE_96		7
#define   VT1732_CODEC_RATE_44		8
#define   VT1732_CODEC_RATE_22		9
#define   VT1732_CODEC_RATE_11		10
#define   VT1732_CODEC_RATE_88		11
#define   VT1732_CODEC_RATE_176		12
#define   VT1732_CODEC_RATE_192		14
#define   VT1732_CODEC_RATE_64		15

#define VT1732_INT_CONFIG1		0x023
#define   VT1732_AC97_2ND_ID		(3 << 6)
#define   VT1732_I2S_MCLK_128		(1 << 3)
#define   VT1732_I2S_SCLK_48		(1 << 2)
#define   VT1732_DATA_FORMAT		(3 << 0)
#define   VT1732_DATA_FMT_I2S		(0 << 0)

#define VT1732_AC97_INDEX		0x024

#define VT1732_AC97_STATUS		0x025
#define   VT1732_AC97_COLD_RESET	(1 << 7)
#define   VT1732_AC97_WARM_RESET	(1 << 6)
#define   VT1732_AC97_WRITE		(1 << 5)
#define   VT1732_AC97_READ		(1 << 4)
#define   VT1732_AC97_READY		(1 << 3)
#define   VT1732_AC97_CODEC_ID		(7 << 0)

#define VT1732_AC97_DATA		0x026

#define VT1732_CONFIG			0x028
#define   VT1732_CFG_CLK192		(1 << 7)
#define   VT1732_CFG_UART2		(1 << 6)
#define   VT1732_CFG_ADC_MASK		(7 << 3)
#define   VT1732_CFG_DAC_MASK		(7 << 0)

#define VT1732_INT_CONFIG		0x029
#define   VT1732_CFG_I2S		(1 << 7)
#define   VT1732_CFG_AC97_SWAP		(1 << 3)
#define   VT1732_CFG_AC97_EXSPLT	(1 << 2)
#define   VT1732_CFG_AC97_SPLIT		(1 << 1)
#define   VT1732_CFG_AC97_SPEC22	(1 << 0)

#define VT1732_I2S_CONFIG		0x02a
#define   VT1732_CFG_I2S_VOLUME		(1 << 7)
#define   VT1732_CFG_I2S_96K		(1 << 6)
#define   VT1732_CFG_I2S_RESMASK	(3 << 5)
#define   VT1732_CFG_I2S_FMT_16		(0 << 5)
#define   VT1732_CFG_I2S_FMT_18		(1 << 5)
#define   VT1732_CFG_I2S_FMT_20		(2 << 5)
#define   VT1732_CFG_I2S_FMT_24		(3 << 5)
#define   VT1732_CFG_I2S_192K		(1 << 3)
#define   VT1732_CFG_I2S_ID_MASK	(3 << 0)

#define VT1732_SPDIF_CONFIG		0x02b
#define   VT1732_CFG_SPDF_EN_TX		(1 << 7)
#define   VT1732_CFG_SPDF_INT_TX	(1 << 6)
#define   VT1732_CFG_SPDF_CLEAR_TX	(1 << 5)
#define   VT1732_CFG_SPDF_EXT_RX	(1 << 1)
#define   VT1732_CFG_SPDF_EXT_TX	(1 << 0)

#define VT1732_SPDF_TX_CTRL		0x02c
#define   VT1732_SPDF_TX_VALID		(1 << 15)
#define   VT1732_SPDF_TX_RATE		(7 << 12)
#define   VT1732_SPDF_TX_RATE_SHIFT	12
#define   VT1732_SPDF_TX_RATE_44	0
#define   VT1732_SPDF_TX_RATE_48	2
#define   VT1732_SPDF_TX_RATE_32	3
#define   VT1732_SPDF_TX_RATE_88	4
#define   VT1732_SPDF_TX_RATE_96	5
#define   VT1732_SPDF_TX_RATE_192	6
#define   VT1732_SPDF_TX_RATE_176	7
#define   VT1732_SPDF_TX_GENLVL		(1 << 11)
#define   VT1732_SPDF_TX_CATEGORY	(0x3f << 4)
#define   VT1732_SPDF_TX_CATEGORY_SHIFT	4
#define   VT1732_SPDF_TX_PREEMP		(1 << 3)
#define   VT1732_SPDF_TX_COPY		(1 << 2)
#define   VT1732_SPDF_TX_NONAUDIO	(1 << 1)
#define   VT1732_SPDF_TX_PROF		(1 << 0)

#define VT1732_PDMA0_BASE_ADDR		0x030	/* up to 0x04c */
#define VT1732_PDMA0_BASE_COUNT		0x050	/* up to 0x05e */
#define VT1732_PDMA0_TERM_COUNT		0x060
#define VT1732_PDMA0_SAMPLE_POS		0x064
#define VT1732_PDMA0_FORMAT1		0x06c
#define   VT1732_PDMA0_SPDF_EN		(1 << 7)
#define   VT1732_PDMA0_CHANNELS_MASK	(15 << 0)
#define VT1732_PDMA0_FORMAT2		0x06f
#define   VT1732_PDMA0_BUFFERS_SHIFT	3
#define   VT1732_PDMA0_FMT_MASK		(3 << 1)
#define   VT1732_PDMA0_FMT_24BIT	(0 << 1)
#define   VT1732_PDMA0_FMT_16BIT	(2 << 1)
#define   VT1732_PDMA0_FMT_8BIT		(3 << 1)
/* offsets */
#define   VT1732_xDMA0_BASE_COUNT_OFS	0x20
#define   VT1732_xDMA0_TERM_COUNT_OFS	0x30
#define   VT1732_xDMA0_SAMPLE_POS_OFS	0x34
#define   VT1732_xDMA0_FORMAT1_OFS	0x3c
#define   VT1732_xDMA0_FORMAT2_OFS	0x3f

#define VT1732_PDMA1_BASE_ADDR		0x070
/* 0x080 - 0x0db: PDMA2-7 */
#define VT1732_PDMASP0_BASE_ADDR	0x0e0
#define VT1732_PDMASP1_BASE_ADDR	0x0f0
#define VT1732_PDMA8_BASE_ADDR		0x100
/* 0x110 - 0x17b: PDMA9-15 */

/* PDMA1-15: offsets */
#define VT1732_PDMAx_BASE_ADDR		0x00
#define VT1732_PDMAx_BASE_COUNT		0x04
#define   VT1732_PDMAx_READ_BUF2	(1 << 16)
#define VT1732_PDMAx_TERM_COUNT		0x08
#define VT1732_PDMAx_FORMAT		0x0b
#define   VT1732_PDMAx_2BUFS		(1 << 3)
#define   VT1732_PDMAx_FMT_MASK		(3 << 1)
#define   VT1732_PDMAx_FMT_24BIT	(0 << 1)
#define   VT1732_PDMAx_FMT_16BIT	(2 << 1)
#define   VT1732_PDMAx_FMT_8BIT		(3 << 1)
#define   VT1732_PDMAx_FMT_STEREO	(1 << 0)

#define VT1732_DMA_FIFO_UNDERRUN	0x1d0
  /* bit0-17: DMA0-7, DMA-SP0&1, DMA8-15 */
#define VT1732_DMA_FIFO_OVERRUN		0x1d4
  /* bit0-19: DMA0-7, DMA-SP0&1, DMA8-15, DMA-DM0&1 */

#define VT1732_GPIO_WRITE_MASK		0x1e0 /* bit0-9 */
#define VT1732_GPIO_DATA		0x1e4 /* bit0-9 */
#define VT1732_GPIO_DATA_DIR		0x1e8 /* bit0-9 */
#define VT1732_GPIO_INT			0x1ec
#define   VT1732_GPIO_INT_CAP		(15 << 8)
#define   VT1732_GPIO_INT_MASK		(15 << 4)
#define   VT1732_GPIO_INT_STATUS	(15 << 0)

#define VT1732_2ND_AC97_INDEX		0x204
#define VT1732_2ND_AC97_STATUS		0x205
#define VT1732_2ND_AC97_DATA		0x206

#define VT1732_RDMA0_BASE_ADDR		0x230	/* up to 0x24c */
#define VT1732_RDMA0_BASE_COUNT		0x250	/* up to 0x25e */
#define VT1732_RDMA0_TERM_COUNT		0x260
#define VT1732_RDMA0_SAMPLE_POS		0x264
#define VT1732_RDMA0_FORMAT1		0x26c
#define   VT1732_RDMA0_SPDF_EN		(1 << 7)
#define   VT1732_RDMA0_MIXER_EN		(1 << 6)
#define   VT1732_RDMA0_CHANNELS_MASK	(15 << 0)

#define VT1732_RDMA1_BASE_ADDR		0x270
/* 0x280 - 0x2db: RMDA0-7 */
#define VT1732_RDMASP0_BASE_ADDR	0x2e0
#define VT1732_RDMASP1_BASE_ADDR	0x2f0
#define VT1732_RDMA8_BASE_ADDR		0x300
/* 0x310 - 0x37b: RMDA9-15 */
#define VT1732_RDMADM0_BASE_ADDR	0x380
#define VT1732_RDMADM0_BASE_COUNT	0x384
#define VT1732_RDMADM0_TERM_COUNT	0x388
#define VT1732_RDMADM0_FORMAT		0x38b
#define VT1732_RDMADM0_SAMPLE_POS	0x38c
#define VT1732_RMDADM1_BASE_ADDR	0x390

#define VT1732_LOOPBACK_OUT		0x3b0
/* bit 0:7 = PSDOUT0-0 - 3_0 routes
 * bit 8:15 = PSDOUT0-1 - 3_1 routes
 * bit 16:23 = PSDOUT4_0 - 7_0 routes
 * bit 24:31 = PSDOUT4_1 - 7_1 routes
 *   0 = DMAx, 1 = DMIXER, 2 = PSDINx loopback, 3 = SPDIF loopback
 */
#define VT1732_OUT_MIX			0x3b4
/* bit 0-7: PDMA0-7 is mixed with PDMA8-15
 * bit 8-15: PDMAx mixed with 6dB attenuation
 */
#define VT1732_LOOPBACK_SPDIF		0x3b6
/* bit 0-3: SPDIF L/R route
 *   0 = DMA, 1 = DMIXER, 2 = PSDINx loopback, 3 = SPDIF loopback
 * bit 4-5: extended PSDINx loopback left/right (define PSDIN*_0 or PSDIN*_1)
 * bit 8-10: PSDIN0-7 for left
 * bit 11: SPDIF-in left=0, right=1
 * bit 12-14: PSDIN0-7 for right
 * bit 15: SPDIF-in left=0, right=1
 */
#define VT1732_PSDOUT0_0_SRC		0x3c0
/* bit 0-3: PSDIN0_0, 0_1, 1_0, ... 7_1
 * bit 7: SPDIF-in left=0, right=1
 */
#define VT1732_PSDOUT0_1_SRC		0x3c1
/* up to 0x3cf: PSDOUT7_1_SRC */

#define VT1732_VOL_CTRL_DATA		0x3d0
#define   VT1732_VOL_CTRL_MASK		0x7f
#define   VT1732_VOL_CTRL_LEFT_SHIFT	0
#define   VT1732_VOL_CTRL_RIGHT_SHIFT	8

#define VT1732_VOL_CTRL_STREAM		0x3d3
/* 0x00-0x14: PDMA-X_0-7,SP0&1, RDMA0-7,SP0&1,DMIXER
 * 0x20-0x27: PDMA-X_8-15
 * 0x4a-0x51: RDMA-X_8-15
 */

#define VT1732_VOL_RATE_CTRL		0x3d4
#define VT1732_VOL_METER_INDEX		0x3d6
/* 0x00-0x0f: PDMA-X_0-15
 * 0x10-0x1f: RDMA-X_0-15
 * 0x20-0x21: PDMA-X_SP0/1
 * 0x22-0x23: RDMA-X-SP0/1
 * 0x24-0x25: RMDA-X-DMIXER0/1
 */
#define VT1732_VOL_METER_DATA		0x3d8

#define VT1732_UART_TX_WATERMARK	0x3e0
#define VT1732_UART_RX_WATERMARK	0x3e1
#define VT1732_UART1_TX_FIFO_STAT	0x3e4
#define VT1732_UART1_RX_FIFO_STAT	0x3e5
#define VT1732_UART2_TX_FIFO_STAT	0x3e6
#define VT1732_UART2_RX_FIFO_STAT	0x3e7
#define VT1732_UART1_DATA		0x3e8
#define VT1732_UART1_STAT		0x3e9
#define VT1732_UART2_DATA		0x3ea
#define VT1732_UART2_STAT		0x3eb

#define VT1732_I2C_DEV_ADDR		0x3f0
#define   VT1732_I2C_WRITE		(1 << 0)
#define VT1732_I2C_BYTE_ADDR		0x3f1
#define VT1732_I2C_DATA			0x3f2
#define VT1732_I2C_STATUS		0x3f3
#define   VT1732_I2C_EEPROM_PRESENT	(1 << 7)
#define   VT1732_I2C_REREAD_EEPROM	(1 << 1)
#define   VT1732_I2C_BUSY		(1 << 0)
#define VT1732_I2C_ENABLE		0x3f4
#define   VT1732_I2C_PORT		(3 << 4)
#define   VT1732_I2C_PORT3_ENABLE	(1 << 3)
#define   VT1732_I2C_PORT2_ENABLE	(1 << 2)
#define   VT1732_I2C_PORT1_ENABLE	(1 << 1)

#define VT1732_SPI_CTRL			0x3f7
#define   VT1732_SPI_ENABLE		(1 << 7)
#define   VT1732_SPI_BUSY		(1 << 4)
#define   VT1732_SPI_DATA_WIDTH		(3 << 2)
#define   VT1732_SPI_DATA_8BIT		(0 << 2)
#define   VT1732_SPI_DATA_16BIT		(1 << 2)
#define   VT1732_SPI_DATA_24IT		(2 << 2)
#define   VT1732_SPI_DATA_32IT		(3 << 2)
#define   VT1732_SPI_CLK_FREQ		(3 << 0)
#define   VT1732_SPI_CLK_4		(0 << 0)
#define   VT1732_SPI_CLK_8		(1 << 0)
#define   VT1732_SPI_CLK_16		(2 << 0)
#define   VT1732_SPI_CLK_32		(3 << 0)
#define VT1732_SPI_DATA			0x3f8

#define VT1732_POWER_DOWN		0x3fc
#define   VT1732_XIN1_CLOCK_DOWN	(1 << 7)
#define   VT1732_I2C_CLOCK_STOP		(1 << 4)
#define   VT1732_MIDI_CLOCK_STOP	(1 << 3)
#define VT1732_EXT_POWER_DOWN		0x3fd
#define   VT1732_XIN2_CLOCK_DOWN	(1 << 7)
#define   VT1732_SPDIF_CLOCK_STOP	(1 << 2)
#define   VT1732_EXT_I2S_CLOCK_STOP	(1 << 0)

#define VT1732_DEBUG			0x3fe

/*
 */

#define vt1732_writeb(chip, reg, data)	writeb(data, (chip)->iomap + (reg))
#define vt1732_writew(chip, reg, data)	writew(data, (chip)->iomap + (reg))
#define vt1732_writel(chip, reg, data)	writel(data, (chip)->iomap + (reg))
#define vt1732_readb(chip, reg)		readb((chip)->iomap + (reg))
#define vt1732_readw(chip, reg)		readw((chip)->iomap + (reg))
#define vt1732_readl(chip, reg)		readl((chip)->iomap + (reg))


/*
 *  
 */

struct snd_vt1732;

enum { EEPROM_CONFIG, EEPROM_IFACE, EEPROM_I2S, EEPROM_SPDIF };

struct snd_vt1732_eeprom {
	unsigned int subvendor;	/* PCI[2c-2f] */
	unsigned char data[4];
};

enum {
	VT1732_PCM_ANALOG0,
	VT1732_PCM_ANALOG15 = 15,
	VT1732_PCM_SPDIF,
	VT1732_PCM_MIXER,
	VT1732_PCM_STREAMS
};

enum {
	MONITOR_AOUT_0,
	MONITOR_AOUT_15 = (MONITOR_AOUT_0 + 15),
	MONITOR_AIN_0,
	MONITOR_AIN_15 = (MONITOR_AIN_0 + 15),
	MONITOR_SPOUT_0, MONITOR_SPOUT_1,
	MONITOR_SPIN_0, MONITOR_SPIN_1,
	MONITOR_MASTER,
	MONITOR_VOLS
};

enum {
	VOLMETER_AOUT_0,
	VOLMETER_AOUT_15 = (VOLMETER_AOUT_0 + 15),
	VOLMETER_AIN_0,
	VOLMETER_AIN_15 = (VOLMETER_AIN_0 + 15),
	VOLMETER_SPOUT_0, VOLMETER_SPOUT_1,
	VOLMETER_SPIN_0, VOLMETER_SPIN_1,
	VOLMETER_MASTER_0, VOLMETER_MASTER_1,
	VOLMETER_NUMS
};


struct snd_vt1732_card_info {
	unsigned int subvendor;
	const char *name;
	const char *model;
	const char *driver;
	const char *shortname;
	unsigned char eeprom[4];
	int (*init)(struct snd_vt1732 *chip);
	void (*free)(struct snd_vt1732 *chip);
	int (*build_controls)(struct snd_vt1732 *chip);
	unsigned int no_mpu401: 1;
	const char *midi_name;
	const char *midi2_name;
};

struct snd_vt1732_stream {
	struct snd_pcm_substream *substream;
	unsigned int buffer_bytes;
	unsigned int nbufs;
	unsigned int running;
	unsigned int base_reg;
	unsigned int stat_bit;
};

struct snd_vt1732 {
	int irq;
	long addr;
	void __iomem *iomap;

	struct pci_dev *pci;
	struct snd_card *card;

	const struct snd_vt1732_card_info *config;

	struct snd_pcm *pcm_interleave;
	struct snd_pcm *pcm_stereo;
	struct snd_pcm *pcm_mono;
	struct snd_pcm *pcm_spdif;
	struct snd_pcm *pcm_mixer;

	struct snd_vt1732_stream str[2][VT1732_PCM_STREAMS];

	unsigned short monitor_vol[MONITOR_VOLS];

	unsigned int rate_locked: 1;
	unsigned int spdif_in_sync: 1;
	int nopened;
	unsigned int rate;
	unsigned int rate_max;

	unsigned int hw_nstreams[2];
	unsigned int spdif_out: 1;
	unsigned int spdif_in: 1;

	unsigned int has_midi: 1;
	unsigned int has_midi2: 1;
	struct snd_rawmidi *rmidi[2];
	struct snd_mpu401 *uart[2];

	spinlock_t reg_lock;
	struct mutex open_mutex;

	struct snd_vt1732_eeprom eeprom;

	struct snd_info_entry *proc_entry;
};

#endif /* __SOUND_VT1732_H */
